/* Snowlog - An Apache logfile browser/analyzer. 
 * 
 * Copyright 2004 Oliver Feiler <kiza@kcore.de>
 * http://snowlog.kcore.de/
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as   
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <strings.h>
#include <stdio.h>
#include <ncurses.h>

#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "regexp.h"
#include "snowtk.h"
#include "internals.h"

struct filter ** filters = NULL;
int count_filters = 0;

int AddFilter (char * name);
int AddFilterRule (struct filter * filters, char * filterline);
filter_type GetFilterType (char * filter);
int CompareRegexp (struct regexp * regexp, const char * string, filter_operator operator);
int CompareIntegers (int one, int two, filter_operator operator);

void LoadFilters (char * filename) {
	int len;
	FILE *file;
	char buf[4096];
	char *name;
	int filter = 0;
	
	UIStatus("Loading filters...", 0, 0, LINES-1);
	file = fopen (filename, "r");
	if (file == NULL)
		return;

	while (!feof(file)) {
		if (fgets(buf, sizeof(buf), file) == NULL)
			break;
		if (buf[0] == '#')
			continue;
		
		len = strlen(buf);
		/* Eat trailing newline */
		if (buf[len-1] == '\n')
			buf[len-1] = 0;

		if (buf[0] == '\0')
			filter = 0;
		
		len = strlen(buf);
		
		if ((buf[0] == '[') && (len>1) && (filter == 0)) {
			name = malloc(len);
			memcpy(name, buf+1, len-2);
			name[len-2] = '\0';
					
			AddFilter(name);
			
			filter = 1;
			continue;
		}
		
		if (filter) {
			AddFilterRule(filters[count_filters-1], buf);
		}
	}
	fclose(file);
}

int AddFilter (char * name) {
	struct filter * newfilter;
	
	newfilter = NewFilterStruct();
	newfilter->name = name;
	if (count_filters % 100 == 0)
		filters = realloc (filters, sizeof(struct filter *)*(count_filters+100));
	count_filters++;
	filters[count_filters-1] = newfilter;
	
	return 0;
}

int AddFilterRule (struct filter * filter, char * filterline) {
	struct filter_rule * rule;
	char *option, *value;
	filter_type type;
	filter_operator operator;
	pcre *compiled;
	pcre_extra *studied;
	int pattern_count;
	struct regexp * regexp;
	
	value = strdup(filterline);
	option = strsep(&value, "\t");
	if (!value) {
		free(option);
		return -1;
	}
	if (value[0] == '=')
		operator = FILTER_EQUALS;
	else if (value[0] == '<')
		operator = FILTER_LESS_THAN;
	else if (value[0] == '>')
		operator = FILTER_GREATER_THAN;
	else if (value[0] == '!')
		operator = FILTER_ISNOT;
	else {
		free(option);
		return -1;
	}
	value++;
	
	if (filter->num_filters % 100 == 0)
		filter->filter_rules = realloc (filter->filter_rules, sizeof(struct filter_rule *)*(filter->num_filters+100));
	rule = malloc(sizeof(struct filter_rule));
	filter->num_filters++;
	
	type = GetFilterType(option);
	
	switch (type) {
		case FILTER_HTTPSTATUS:
		case FILTER_CONTENT_LENGTH:
			rule->type = type;
			rule->operator = operator;
			rule->filter = malloc(sizeof(int));
			*((int *)rule->filter) = atoi(value);
			break;
		case FILTER_METHOD:
		case FILTER_REQUEST:
		case FILTER_MIME_TYPE:
		case FILTER_REFERER:
		case FILTER_USERAGENT:
		case FILTER_VHOST:
		case FILTER_AUTHNAME:
		case FILTER_LOGHINT:
			if ((operator == FILTER_GREATER_THAN) || (operator == FILTER_LESS_THAN)) {
				/* Invalid filter defintion */
				free(option);
				free(rule);
				filter->num_filters--;
				return -1;
			}
			rule->type = type;
			rule->operator = operator;
			compiled = NewCompiledRegexp(value);
			if (!compiled) {
				free(option);
				free(rule);
				filter->num_filters--;
				return -1;
			}
			studied = NewStudiedRegexp(compiled);
			pattern_count = PatternCountInRegexp(compiled, studied);
			regexp = NewRegexpStruct();
			regexp->regexp = strdup(value);
			regexp->compiled = compiled;
			regexp->studied = studied;
			regexp->regexp_pattern_count = pattern_count;
			rule->filter = regexp;
			break;
		case FILTER_NOBOTS:
		case FILTER_NONE:
		default:
			free(option);
			free(rule);
			filter->num_filters--;
			return -1;
			break;
	}
	
	filter->filter_rules[filter->num_filters-1] = rule;
	
	free(option);
	
	return 0;
}


filter_type GetFilterType (char * filter) {
	if (strcasecmp(filter, "httpstatus") == 0)
		return FILTER_HTTPSTATUS;
	else if (strcasecmp(filter, "content_length") == 0)
		return FILTER_CONTENT_LENGTH;
	else if (strcasecmp(filter, "method") == 0)
		return FILTER_METHOD;
	else if (strcasecmp(filter, "request") == 0)
		return FILTER_REQUEST;
	else if (strcasecmp(filter, "mime_type") == 0)
		return FILTER_MIME_TYPE;
	else if (strcasecmp(filter, "referer") == 0)
		return FILTER_REFERER;
	else if (strcasecmp(filter, "useragent") == 0)
		return FILTER_USERAGENT;
	else if (strcasecmp(filter, "vhost") == 0)
		return FILTER_VHOST;
	else if (strcasecmp(filter, "authname") == 0)
		return FILTER_AUTHNAME;
	else if (strcasecmp(filter, "loghint") == 0)
		return FILTER_LOGHINT;
	else if (strcasecmp(filter, "bots") == 0)
		return FILTER_NOBOTS;
	else
		return FILTER_NONE;
}

void PrintFilters (void) {
	int i, j;
	
	printf("\t\ttype\toperator\tstring\n\n");
	for (i = 0; i < count_filters; i++) {
		printf("Filter: %s\n", filters[i]->name);
		for (j = 0; j < filters[i]->num_filters; j++) {
			printf("\t\t%d\t%d\t", filters[i]->filter_rules[j]->type,
								   filters[i]->filter_rules[j]->operator);
			switch (filters[i]->filter_rules[j]->type) {
				case FILTER_HTTPSTATUS:
				case FILTER_CONTENT_LENGTH:
					printf("%d\n", *((int *)filters[i]->filter_rules[j]->filter));
					break;
				case FILTER_METHOD:
				case FILTER_REQUEST:
				case FILTER_MIME_TYPE:
				case FILTER_REFERER:
				case FILTER_USERAGENT:
				case FILTER_VHOST:
				case FILTER_AUTHNAME:
				case FILTER_LOGHINT:
					printf("%s\n", ((struct regexp *)filters[i]->filter_rules[j]->filter)->regexp);
					break;
				default:
					printf("FUCKED!\n");
			}
		}
	}
}


int ItemMatchesFilter (int filter_to_use, struct logentry * item) {
	int i, rc;
	filter_operator operator;
	struct filter_rule * rule;
		
	for (i = 0; i < filters[filter_to_use-1]->num_filters; i++) {
		operator = filters[filter_to_use-1]->filter_rules[i]->operator;
		rule = filters[filter_to_use-1]->filter_rules[i];
		
		switch (filters[filter_to_use-1]->filter_rules[i]->type) {
			case FILTER_HTTPSTATUS:
				rc = CompareIntegers(item->httpstatus,
									 *((int *)filters[filter_to_use-1]->filter_rules[i]->filter),
									 operator);
				if (rc == 0)
					return 0;
				break;
			case FILTER_CONTENT_LENGTH:
				rc = CompareIntegers(item->content_length,
									 *((int *)filters[filter_to_use-1]->filter_rules[i]->filter),
									 operator);
				if (rc == 0)
					return 0;
				break;
			case FILTER_METHOD:
				rc = CompareRegexp(rule->filter, item->method, operator);
				if (rc == 0)
					return 0;
				break;
			case FILTER_REQUEST:
				rc = CompareRegexp(rule->filter, item->request, operator);	
				if (rc == 0)
					return 0;
				break;
			case FILTER_MIME_TYPE:
				rc = CompareRegexp(rule->filter, item->mime_type, operator);
				if (rc == 0)
					return 0;
				break;
			case FILTER_REFERER:
				rc = CompareRegexp(rule->filter, item->referer, operator);
				if (rc == 0)
					return 0;
				break;
			case FILTER_USERAGENT:
				rc = CompareRegexp(rule->filter, item->useragent, operator);
				if (rc == 0)
					return 0;
				break;
			case FILTER_VHOST:
				rc = CompareRegexp(rule->filter, item->vhost, operator);
				if (rc == 0)
					return 0;
				break;
			case FILTER_AUTHNAME:
				rc = CompareRegexp(rule->filter, item->authname, operator);
				if (rc == 0)
					return 0;
				break;
			case FILTER_LOGHINT:
				rc = CompareRegexp(rule->filter, item->loghint, operator);
				if (rc == 0)
					return 0;
				break;
			default:
				return 0;
				break;
		}
	}
	
	return 1;
}

int CompareRegexp (struct regexp * regexp, const char * string, filter_operator operator) {
	int rc;
	
	rc = MatchRegexp(regexp, string);
	
	switch (operator) {
		case FILTER_EQUALS:
			if (rc == 1)
				return 1;
			else
				return 0;
			break;
		case FILTER_ISNOT:
			/* Negated */
			if (rc == 1)
				return 0;
			else
				return 1;
			break;
		case FILTER_GREATER_THAN:
		case FILTER_LESS_THAN:
			/* Doesn't make sense */
		default:
			return 0;
			break;
	}
	return 0;
}

int CompareIntegers (int one, int two, filter_operator operator) {
	switch (operator) {
		case FILTER_EQUALS:
			if (one == two)
				return 1;
			else
				return 0;
		case FILTER_GREATER_THAN:
			if (one > two)
				return 1;
			else
				return 0;
		case FILTER_LESS_THAN:
			if (one < two)
				return 1;
			else
				return 0;
		case FILTER_ISNOT:
			if (one != two)
				return 1;
			else
				return 0;
		default:
			return 0;
			break;
	}
	return 0;
}

