/* Snowlog - An Apache logfile browser/analyzer.
 * 
 * Copyright 2004 Oliver Feiler <kiza@kcore.de>
 * http://snowlog.kcore.de/
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as   
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */
   
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>

#include <ncurses.h>

#include "config.h"

#include "snowtk.h"
#include "internals.h"
#include "filters.h"
#include "regexp.h"

extern int count_requests, count_hosts;
int first_date = 2147483647;		/* Ok, this hack is really crappy! And will fail in 2038! */
int last_date = 0;

struct logentry * NewLogentryStruct (void) {
	struct logentry * new;
	
	new = malloc (sizeof(struct logentry));

	new->httpversion = 0;
	new->httpstatus = 0;
	new->content_length = 0;
	new->timestamp = 0;
	new->method = NULL;
	new->request = NULL;
	new->mime_type = NULL;
	new->referer = NULL;
	new->useragent = NULL;
	new->vhost = NULL;
	new->authname = NULL;
	new->loghint = NULL;
	new->isbot = 0;
	
	return new;
}

struct host * NewHostStruct (void) {
	struct host * new;
	
	new = malloc(sizeof(struct host));
	
	new->ipaddr = NULL;
	new->hostname = NULL;
	new->first_referer = NULL;
	new->smart_first_referer = NULL;
	new->entries = NULL;
	new->num_entries = 0;
	new->total_bytes = 0;
	
	return new;
}

struct displayitem * NewDisplayItemStruct (int columns) {
	struct displayitem * new;
	
	new = malloc(sizeof(struct displayitem));
	
	new->skip = 0;
	new->indent = 0;
	new->columns = malloc(sizeof(char *)*(columns+1));
	memset (new->columns, 0, sizeof(char *)*(columns+1));
	new->object = NULL;
	
	return new;
}

struct displaystats * NewDisplayStatsStruct (void) {
	struct displaystats * new;
	
	new = malloc(sizeof(struct displaystats));
	
	new->count_hosts = 0;
	new->count_requests = 0;
	new->count_bytes = 0;
	
	return new;
}

struct regexp * NewRegexpStruct (void) {
	struct regexp * new;
	
	new = malloc(sizeof(struct regexp));
	
	new->regexp = NULL;
	new->substitution = NULL;
	new->compiled = NULL;
	new->studied = NULL;
	
	return new;
}

struct filter * NewFilterStruct (void) {
	struct filter * new;
	
	new = malloc(sizeof(struct filter));
	
	new->name = NULL;
	new->num_filters = 0;
	new->filter_rules = NULL;
	
	return new;
}

struct filter_rule * NewFilterRuleStruct (void) {
	struct filter_rule * new;
	
	new = malloc(sizeof(struct filter_rule));
	
	new->filter = NULL;
	
	return new;
}

struct displayitem ** DisplayItemsFromHosts (struct host ** hosts,
											 struct displaystats * stats,
											 int filter_to_use) {
	int i, j, count, host_added;
	char httpstatus_string[4];
	struct displayitem **items = NULL;
	
	stats->count_hosts = 0;
	stats->count_requests = 0;
	stats->count_bytes = 0;
	
	UIStatus ("Snowlog thinking...", 0, 0, LINES-1);

	if (count_hosts <= 0)
		return NULL;
	
	items = malloc(sizeof(struct displayitem *)*(count_hosts+count_requests));
	count = 0;
	
	for (i = 0; i < count_hosts; i++) {
		host_added = 0;
		for (j = 0; j < hosts[i]->num_entries; j++) {
			if (filter_to_use) {
				if (!ItemMatchesFilter(filter_to_use, hosts[i]->entries[j]))
					continue;
			}
			
			if (host_added != 1) {
				/* Only add a host entry if at least one request matched */
				items[count] = NewDisplayItemStruct(2);
				items[count]->skip = 1;
				items[count]->indent = 0;
				items[count]->type = ENTRY_TYPE_HOST;
				items[count]->num_columns = 2;
				items[count]->columns[0] = hosts[i]->ipaddr;
				items[count]->columns[1] = hosts[i]->smart_first_referer;
				items[count]->object = hosts[i];
				count++;
				stats->count_hosts++;
				host_added = 1;
			}

			/* Add request entry */
			snprintf (httpstatus_string, sizeof(httpstatus_string),
					  "%d", hosts[i]->entries[j]->httpstatus);
			items[count] = NewDisplayItemStruct(2);
			items[count]->skip = 0;
			items[count]->indent = 1;
			items[count]->type = ENTRY_TYPE_REQUEST;
			items[count]->num_columns = 2;
			items[count]->columns[0] = hosts[i]->entries[j]->request;
			items[count]->columns[1] = strdup(httpstatus_string);
			items[count]->object = hosts[i]->entries[j];
			count++;
			
			/* Build the stats tables */
			if (hosts[i]->entries[j]->timestamp < first_date)
				first_date = hosts[i]->entries[j]->timestamp;
			if (hosts[i]->entries[j]->timestamp > last_date)
				last_date = hosts[i]->entries[j]->timestamp;
			
			stats->count_requests++;
			stats->count_bytes += hosts[i]->entries[j]->content_length;
		}
	}
	if (count == 0) {
		free (items);
		items = NULL;
	}
	return items;
}

void FreeDisplayList (struct displayitem ** list, int num_items) {
	int i;
	
	for (i = 0; i < num_items; i++) {
		if (list[i]->type == ENTRY_TYPE_REQUEST)
			free((char *) list[i]->columns[1]);
		free(list[i]->columns);
		free(list[i]);
	}
	free(list);
}

void StringWithSmartBytes (char * string, int string_size, unsigned long long bytes) {
	double newsize;
	
	if (bytes > (unsigned long long)1024*1024*1024*1024*1024*1024) {
		newsize = (double)bytes / ((double)(unsigned long long)1024*1024*1024*1024*1024*1024);
		snprintf (string, string_size, "%.4f EB", newsize);
	} else if (bytes > (unsigned long long)1024*1024*1024*1024*1024) {
		newsize = (double)bytes / ((double)(unsigned long long)1024*1024*1024*1024*1024);
		snprintf (string, string_size, "%.4f PB", newsize);
	} else if (bytes > (unsigned long long)1024*1024*1024*1024) {
		newsize = (double)bytes / ((double)(unsigned long long)1024*1024*1024*1024);
		snprintf (string, string_size, "%.3f TB", newsize);
	} else if (bytes > 1024*1024*1024) {
		newsize = (double)bytes / ((double)1024*1024*1024);
		snprintf (string, string_size, "%.2f GB", newsize);
	} else if (bytes > 1024*1024) {
		newsize = (double)bytes / ((double)1024*1024);
		snprintf (string, string_size, "%.2f MB", newsize);
	} else if (bytes > 1024) {
		newsize = bytes / 1024;
		snprintf (string, string_size, "%d KB", (int)newsize);
	} else {
		snprintf (string, string_size, "%llu B", bytes);
	}
}

void ReplaceCharFromString (char * string, char remove, char replace) {
	int i, len;
	
	if (string == NULL)
		return;
	
	len = strlen(string);
	for (i = 0; i < len; i++) {
		if (string[i] == remove)
			string[i] = replace;
	}
}

char * URLDecodeString (char * string) {
	int i = 0, j = 0, k;
	int len;
	char *encoded;
	char hexchar[3];
	char newchar;
	char *newstring;
	
	if (string == NULL)
		return strdup("(Greetings from the depths of the memory manager!)");
	
	len = strlen(string);
	newstring = malloc(len+1);
	for (;;) {
		if (string[i] == 0) {
			newstring[j] = 0;
			break;
		}
		if (string[i] == '%') {
			encoded = string+(i+1);
			memcpy(hexchar, encoded, 2);
			hexchar[2] = 0;
			/* Only A-F need to be converted to lowercase. */
			for (k = 0; k < 2; k++) {
				if ((hexchar[k] >= 65) && (hexchar[k] <= 70))
					hexchar[k] += 32;
			}
			/* We ignore anything else and set newchar to an invalid value. */
			if (sscanf(hexchar, "%hhX", &newchar) == 0)
				newchar = 0;
			if ((newchar >= 32) && (newchar <= 126)) {
				newstring[j] = newchar;
				i += 3;
				j++;
				continue;
			}
			/* else just leave it alone! */
		}
		newstring[j] = string[i];
		i++;
		j++;
	}
	return newstring;
}
