/* Snowlog - An Apache logfile browser/analyzer.
 * 
 * Copyright 2004 Oliver Feiler <kiza@kcore.de>
 * http://snowlog.kcore.de/
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as   
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */
   
#include <ncurses.h>
#include <pcre.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <assert.h>

#include "config.h"
#include "internals.h"
#include "snowtk.h"

pcre * NewCompiledRegexp (char * regexp);
pcre_extra * NewStudiedRegexp (const pcre * compiled);
int PatternCountInRegexp (pcre * compiled, pcre_extra * studied);
int AddToRegexpArray (struct regexp *** array, int * count,
					  char * regexp, char * substitution,
					  pcre * compiled, pcre_extra * studied,
					  int regexp_pattern_count);
char * NewSubstitutedString (const char ** listptr, const char * subststring,
							 int substitutions);

int LoadRegexps (char * filename, struct regexp *** array, int * count) {
	char buf[4096];
	char *regexp, *substitution;
	FILE * regexp_file;
	pcre *compiled;
	pcre_extra *studied;
	int regexp_pattern_count;
	int len;
	
	UIStatus("Loading regular expressions...", 0, 0, LINES-1);
	regexp_file = fopen(filename, "r");
	if (regexp_file == NULL)
		return -1;
	
	while (!feof(regexp_file)) {
		if (fgets(buf, sizeof(buf), regexp_file) == NULL)
			break;
		if ((buf[0] == '#') || (buf[0] == '\n'))
			continue;

		/* Eat trailing newline */
		len = strlen(buf);
		if (buf[len-1] == '\n')
			buf[len-1] = 0;
				
		substitution = strdup(buf);
		regexp = strsep(&substitution, "\t");
		
		compiled = NewCompiledRegexp(regexp);
		if (!compiled) {
			free(regexp);
			continue;
		}
		studied = NewStudiedRegexp(compiled);
		regexp_pattern_count = PatternCountInRegexp(compiled, studied);
		if (regexp_pattern_count > 0)
			AddToRegexpArray(array, count, regexp, substitution,
						     compiled, studied, regexp_pattern_count);
		else {
			free(regexp);
			free(compiled);
			free(studied);
		}
	}
	fclose(regexp_file);
	
	return 0;
}

pcre * NewCompiledRegexp (char * regexp) {
	pcre *compiled;
	const char *err;
	int erroffset;
	
	compiled = pcre_compile(regexp, 0, &err, &erroffset, NULL);

	return compiled;
}

pcre_extra * NewStudiedRegexp (const pcre * compiled) {
	pcre_extra *studied;
	const char *err;
	
	studied = pcre_study(compiled, 0, &err);
	
	return studied;
}

int PatternCountInRegexp (pcre * compiled, pcre_extra * studied) {
	int rc;
	unsigned long int len;
	
	rc = pcre_fullinfo(compiled, studied, PCRE_INFO_CAPTURECOUNT, &len);
	
	if (rc == 0)
		return len;
	else
		return -1;
}

int AddToRegexpArray (struct regexp *** array, int * count,
					  char * regexp, char * substitution,
					  pcre * compiled, pcre_extra * studied,
					  int regexp_pattern_count) {
	struct regexp *newregexp;
	
	if (*count % 100 == 0)
		*array = realloc (*array, sizeof(struct regexp *)*(*count+100));

	(*count)++;
	
	newregexp = NewRegexpStruct();
	newregexp->regexp = regexp;
	newregexp->substitution = substitution;
	newregexp->compiled = compiled;
	newregexp->studied = studied;
	newregexp->regexp_pattern_count = regexp_pattern_count;
	(*array)[*count-1] = newregexp;
	
	return 0;
}

int MatchRegexp (struct regexp * array, const char * string) {
	int rc, ovecsize, *ovector;

	ovecsize = (array->regexp_pattern_count+1)*3;
	ovector = malloc(sizeof(int)*ovecsize);	
	rc = pcre_exec(array->compiled,
				   array->studied,
				   string,
				   strlen(string),
				   0,
				   0,
				   ovector,
				   ovecsize);
	free(ovector);
	assert(rc);
	if (rc > 0)
		return 1;
	else return 0;
}

char * NewRegexpMatchedString (struct regexp ** array, int array_count,
							   const char * string) {
	char *newstring;
	int *ovector = NULL;
	int ovecsize;
	int i, rc;
	const char **listptr;
	
	rc = 0;
	for (i = 0; i < array_count; i++) {
		ovecsize = (array[i]->regexp_pattern_count+1)*3;
		ovector = malloc(sizeof(int)*ovecsize);
	
		rc = pcre_exec(array[i]->compiled,
					   array[i]->studied,
					   string,
					   strlen(string),
					   0,
					   0,
					   ovector,
					   ovecsize);
		if (rc > 0) {
			/* Match found */
			break;
		}
		free(ovector);
	}
	
	if (rc > 0) {
		pcre_get_substring_list(string, ovector, rc, &listptr);
		newstring = NewSubstitutedString(listptr, array[i]->substitution, rc);
		pcre_free_substring_list(listptr);
		free(ovector);
	} else newstring = strdup(string);
	
	return newstring;
}

char * ReplaceString(const char *haystack, const char *needle, const char *replacestr) {
	/* Returns a newly allocated string where needle in haystack is replaced by
	 * replacestr. The returned string must be freed by the caller.
	 */
	char *needleptr, *newstr;

	needleptr = strstr(haystack, needle);
	if (!needleptr)
		return strdup(haystack);
		
	newstr = malloc(strlen(haystack)-strlen(needle)+strlen(replacestr)+1);
	*newstr = 0;
	strncat(newstr, haystack, needleptr-haystack);
	strcat(newstr, replacestr);
	needleptr += strlen(needle);
	strcat(newstr, needleptr);
	
	return newstr;
}

char * NewSubstitutedString (const char ** listptr, const char * subststring,
							 int substitutions) {
	char *newstring = NULL;
	int i;
	char searchstr[8];
	char *laststring;
	
	newstring = strdup(subststring);
	
	for (i = 1; i < substitutions; i++) {
		snprintf(searchstr, sizeof(searchstr), "$%d", i);
		laststring = newstring;
		newstring = ReplaceString(newstring, searchstr, listptr[i]);
		free(laststring);
	}
	
	return newstring;
}

