/* Snowlog - An Apache logfile browser/analyzer.
 * 
 * Copyright 2004 Oliver Feiler <kiza@kcore.de>
 * http://snowlog.kcore.de/
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as   
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */
   
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/time.h>
#include <unistd.h>

#include <ncurses.h>

#include "config.h"

#include "snowtk.h"
#include "internals.h"
#include "regexp.h"

struct host **hosts = NULL;
int count_hosts = 0;
int count_requests = 0;

extern struct regexp ** useragents;
extern int useragents_count;
extern struct regexp ** referers;
extern int referers_count;

int LogParseEntry (char * buf);
int LogAddToMainlog (char * ipaddr, struct logentry * entry);
int LogAddHostEntry (struct host * log, struct logentry * entry);

static struct snow_stringhash * stringhash = NULL;

int LogLoad (char * filename) {
	char buf[16384];		/* This used to be 4096 until I tried to parse an IIS exploit... */
	int len;
	FILE * logfile;
	char tmp[256];
	struct timeval tv;
	long oldtime;
	
	if (stringhash == NULL) {
		stringhash = snow_stringhash_init(1024);
	}
	
	gettimeofday(&tv, NULL);
	oldtime = tv.tv_sec;
	
	UIStatus("Parsing log, please wait...", 0, 0, LINES-1);
	logfile = fopen (filename, "r");
	if (logfile == NULL)
		return -1;
	
	while (!feof(logfile)) {
		if (fgets(buf, sizeof(buf), logfile) == NULL)
			break;
		/* Eat trailing newline */
		len = strlen(buf);
		if (buf[len-1] == '\n')
			buf[len-1] = 0;
		LogParseEntry (buf);
		
		gettimeofday(&tv, NULL);
		if (tv.tv_sec-oldtime >= 1) {
			snprintf(tmp, sizeof(tmp), "%d entries parsed to far...", count_requests);
			UIStatus(tmp, 0, 0, LINES-1);
			oldtime = tv.tv_sec;
		}
	}
	return 0;
}

/*
LogFormat "
%a	%l	%u	%t	%r	%>s	%X	%{Content-Type}o	%b	%{Referer}i	%{User-Agent}i	%V	%{LogHint}e
" parsable
0	1	2	3	4	5	6	7					8	9			10				11	12

213.23.26.162   -       -       [28/Jun/2004:00:01:05 +0200]    GET /newsweave/rd
fstore.cgi HTTP/1.1     200     +       text/plain      353     -       Mozilla/5
.0 (Macintosh; U; PPC Mac OS X; de-de) AppleWebKit/125.2 (KHTML, like Gecko) Safa
ri/125.8        www.kcore.de    kcore
*/
int LogParseEntry (char * buf) {
	struct logentry * entry;
	char *logline[13];
	char *request[3];
	int i;
	char *ipaddr;
	
	assert(buf);
	
	/* Split one line into its seperate chunks. */
	for (i = 0; i <= 12; i++) {
		logline[i] = strsep (&buf, "\t");
	}
	if (logline[12] == NULL)
		return -1;
	
	/* logline[4] combines HTTP method, path and HTTP protocol. */
	for (i = 0; i <= 2; i++)
		request[i] = strsep (&logline[4], " ");
	
	if (request[2] == NULL)
		return -1;
	
	entry = NewLogentryStruct();
	
	ipaddr = strdup(logline[0]);
	entry->authname = snow_stringhash_unique(stringhash, logline[2]);
	entry->timestamp = atoi(logline[3]);
	entry->method = snow_stringhash_unique(stringhash, request[0]);
	entry->request = snow_stringhash_unique(stringhash, request[1]);
	entry->httpversion = request[2][7]-48;
	entry->httpstatus = atoi(logline[5]);
	entry->mime_type = snow_stringhash_unique(stringhash, logline[7]);
	if (logline[8][0] != '-')
		entry->content_length = strtoll(logline[8], NULL, 10);
	else
		entry->content_length = 0;
	entry->referer = snow_stringhash_unique(stringhash, logline[9]);
	entry->useragent = snow_stringhash_unique(stringhash, logline[10]);
	entry->vhost = snow_stringhash_unique(stringhash, logline[11]);
	entry->loghint = snow_stringhash_unique(stringhash, logline[12]);
	
	LogAddToMainlog (ipaddr, entry);

	free(ipaddr);
	return 0;
}

int LogAddToMainlog (char * ipaddr, struct logentry * entry) {
	struct host *newhost;
	int i;
	
	for (i = 0; i < count_hosts; i++) {
		if (strcmp(ipaddr, hosts[i]->ipaddr) == 0) {
			LogAddHostEntry (hosts[i], entry);
			return 0;
		}
	}
	
	if (count_hosts % 1000 == 0)
		hosts = realloc (hosts, sizeof(struct host *)*(count_hosts+1000));
	count_hosts++;
	
	newhost = NewHostStruct();
	newhost->ipaddr = strdup(ipaddr);
	newhost->entries = NULL;
	newhost->num_entries = 0;
	hosts[count_hosts-1] = newhost;
	LogAddHostEntry (hosts[count_hosts-1], entry);
	
	return 0;
}

int LogAddHostEntry (struct host * host, struct logentry * entry) {
	char * referer_string;
	
	if (host->num_entries % 1000 == 0)
		host->entries = realloc (host->entries, sizeof(struct logentry *)*(host->num_entries+1000));
	host->entries[host->num_entries] = entry;
	host->num_entries++;
	host->total_bytes += entry->content_length;
	count_requests++;
	if (host->num_entries == 1) {
		host->first_referer = strdup(entry->referer);
		referer_string = NewRegexpMatchedString(referers,
		                                        referers_count,
		                                        entry->referer);
		ReplaceCharFromString(referer_string, '+', ' ');
		host->smart_first_referer = URLDecodeString(referer_string);
		free(referer_string);
	}
	
	return 0;
}
